#!/usr/bin/perl
#!/usr/bin/speedy
# ----------------------------------------------------------------------
# Copyright (C) 2005 Mark Lawrence <nomad@null.net>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# ----------------------------------------------------------------------
# parselog.cgi - Generate graphs from parselog RRDs
# ----------------------------------------------------------------------
use strict;
use warnings;
use NCGI;
use RRDs;
use DateTime;
use debug;

our $conf;
init();

# Call this as early as possible
my $cgi = NCGI->instance;

my $defid = 0;
my $cdefid = 0;

my $base = $ENV{SCRIPT_NAME} || '';
my $path = $ENV{PATH_INFO} || shift(@ARGV) || '';
   $path =~ s/^\///g; # remove leading '/'

my @paths  = split('/', $path);
my $mpoint = join(' :: ', @paths);

my $title = "Parselog Statistics";
if (@paths) {
    $title .= " for $mpoint";
}

my $html = $cgi->content;

$html->_goto('head');
$html->title($title);
$html->meta({name => 'generator', content => "parselog-cgi"});
$html->link({
    'rel'   => 'stylesheet',
    'media' => 'screen',
    'href'  => '/parselog-cgi/css/style.css',
    'type'  => 'text/css',
});

$html->_goto('body');
$html->div_open({id => 'header'});



my $query = $cgi->params;
$query->{period} = $query->{period} || '';

if (!exists($conf->{width}->{$query->{period}})) {
    $query->{period} = $conf->{default_period};
}
my ($start,$end) = get_start_end($query->{period});

my @a = split('/', $path);
#shift(@a); # first one is always blank;
foreach my $x ('target', 'mod', 'src') {
    $query->{$x} = shift(@a) || '';
}

#
# Breadcrumbs navigation
#
$html->div_open({id => 'breadcrumbs'});
$html->a({href => $base}, 'Parselog');
if (@paths) {
    my @a = @paths;
    my $last = pop(@a);
    my $p = $base;
    foreach my $x (@a) {
        $html->_add(' :: ');
        $html->a({href => $p.'/'.$x}, $x);
        $p .= "/$x";
    }
    $html->_add(" :: $last");
}
$html->div_close();

debug::log("breadcrumbs completed") if(DEBUG);


#
# Time Period selection
#
$html->form_open({action => "$base/$path", method => 'get'});
$html->div_open({id => 'period'});
$html->select_open({name => 'period'});

my $current_period = $query->{period} || '';
foreach ('today', 'yesterday', 'two days', 'this week', 'last week',
         'this month', 'last month', 'this year', 'last year') {
    (my $val = $_) =~ s/ /_/g;
    my $hash = {value => $val};
    if ($val eq $current_period) {
        $hash->{selected} = 'selected';
    }
    $html->option($hash, $_);
}

$html->select_close();
$html->input({type => 'submit', value => 'Go'});

$html->div_close();
$html->form_close();

debug::log("time period selection completed") if(DEBUG);


$html->div_close(); # header

#$html->pre(Dumper($query));


$html->div_open({id => 'content'}, undef);


chdir($conf->{libdir}) || $html->p("Couldn't chdir($conf->{libdir}): $!");


if (!@paths) {
    foreach my $dir (grep(!/(^_)|(^state$)/, <*>)) {
        $html->p_open;
        $html->a({href => $base.'/'.$dir}, $dir);
        $html->p_close;
    }
    $cgi->respond();
    exit;
}




my @args = (

);


#foreach my $obj ('target', 'mod') {
#    $conf->{$obj} = shift(@paths) || '';
#}

#
# The target of the current query...
#
if (!chdir($conf->{libdir} .'/'. $path)) {
    $html->p("Could not chdir $conf->{libdir}/$path");
    $cgi->respond;
    exit;
};

#
# Obtain a list of all sources for this target
#
my @srcs = (<_*>);
my %types;
foreach my $s (@srcs) {
    chdir($s) || $html->p("Couldn't chdir $s: $!");
    foreach (<*.rrd>) { 
        $types{$_} = $_;
        (my $v1 = $_) =~ s/\.rrd/\.state/;
        if (-f $v1 and open(TMP, $v1)) {
            while (my $line = <TMP>) {
                if ($line =~ /name: (.*)/) {
                    $types{$_} = $1;
                }
            }
            close(TMP);
        }
    }
    chdir('..');
}

#
# Add any directories in this target, that might not have their own RRD
# in the current directory
#
foreach my $s (grep(!/^_/, <*>)) {
    if (!exists($types{$s})) {
        $types{$s.'.rrd'} = $s;
    }
}

foreach my $e (sort keys %types) {
    (my $element = $e) =~ s/\.rrd//;

    my %a = (target => 'target', mod => 'mod');

    my $img = join('-', @paths) . "-$e-$query->{period}.png";

    my @colors = (
        '00ff00','11ff00','22ff00','33ff00','44ff00','55ff00','66ff00',
        '77ff00','88ff00','99ff00','aaff00','bbff00','ccff00','ddff00',
        'eeff00','ffff00','ffee00','ffdd00','ffcc00','ffbb00','ffaa00',
        'ff9900','ff8800','ff7700','ff6600','ff5500','ff4400','ff3300',
        'ff2200','ff1100','ff0000','ff0011','ff0022','ff0033','ff0044',
        'ff0055','ff0066','ff0077','ff0088','ff0099','ff00aa','ff00bb',
        'ff00cc','ff00dd','ff00dd','ff00ee','ff00ff','ee00ff','dd00ff',
        'cc00ff','bb00ff','aa00ff','9900ff','8800ff','7700ff','6600ff',
        '5500ff','4400ff','3300ff','2200ff','1100ff','0000ff',
    );

    my @args = (
        "--title=$mpoint :: $types{$e}",
        '--start='. $start->epoch(),
        '--end='  . $end->epoch(),
        '--width=' . $conf->{width}->{$query->{period}},
        '--rigid',
        '--lower-limit=0',
    );

    my $i = 0;
    my $j = 0;
    my $stack = 'AREA';
    my $dir = 0;

    if (-d $element) {
        $dir = 1;
        chdir($element) || $html->p("Couldn't chdir $element $!");
        my @srcs = (<_*>);
        my %types;
        foreach my $s (@srcs) {
            chdir($s) || $html->p("Couldn't chdir $s: $!");
            foreach (<*.rrd>) { 
                $types{$_} = $_;
                (my $v1 = $_) =~ s/\.rrd/\.state/;
                if (-f $v1 and open(TMP, $v1)) {
                    while (my $line = <TMP>) {
                        if ($line =~ /name: (.*)/) {
                            $types{$_} = $1;
                        }
                    }
                    close(TMP);
                        }
            }
            chdir('..');
        }

        foreach my $e (sort keys %types) {
            my $color = $colors[$j];
            while (my ($str,$col) =  each %{$conf->{color}}) {
                my $tmp = $path.'/'.$element.'/'.$types{$e};
                if ($tmp =~ m!$str!) {
                    $color = $col;
                }
            }
            push(@args, add_target($e, $stack, $color));
            $stack = 'STACK';
            $j++;
        }
    }
    else {
        my $color = $colors[$j];
        while (my ($str,$col) =  each %{$conf->{color}}) {
            my $tmp = $path.'/'.$types{$e};
            if ($tmp =~ m!$str!) {
                $color = $col;
            }
        }
        push(@args, add_target($e, $stack, $color));
        $j++;
    }


#    $html->pre("RRDs::graph: $conf->{imgdir}/$img\n\t", join("\n\t", @args));

    my $err;
    my ($avgs, $xsize, $ysize) = RRDs::graph($conf->{imgdir} .'/'. $img, @args);
    if ($err = RRDs::error) {
        $html->pre('RRDs::graph: ' . $err);
        next;
    }

    if ($dir) {
        $html->a_open({href => "$base/$path/$element?period=$query->{period}"});
    }

    $html->img({
        src    => $conf->{imgurl} .'/'. $img,
        alt    => "$a{mod} Statistics for $a{target}",
        height => $ysize,
        width  => $xsize,
    });

    if ($dir) {
        $html->a_close();
        chdir('..');
    }
    $html->br();
}



$html->div_close();
$html->p({style => 'font-size: 0.7em;'},
         'Generated by <a href="http://rekudos.net/parselog/">parselog</a>.');
$html->body_close();

$cgi->respond();


sub add_target {
    my $rrd   = shift;
    my $stack = shift;
    my $color = shift;
    my $name  = $rrd;

    my @args;
    my @srcs = (<_*>);
    my $cdef = "CDEF:cdef$cdefid=0";
    foreach my $s (@srcs) {
        if (-f "$s/$rrd") {
            (my $state = $rrd) =~ s/\.rrd/\.state/;
            if (-f "$s/$state" and open(TMP, "$s/$state")) {
                while (my $line = <TMP>) {
                    if ($line =~ /name: (.*)/) {
                        ($name = $1) =~ s/:/\\:/g;
                    }
                }
                close(TMP);
            }
            push(@args, "DEF:def$defid=$s/$rrd:hits:AVERAGE");
            $cdef .= ",def$defid,+";
            $defid++;
        }
    }
    push(@args, $cdef, "$stack:cdef$cdefid#$color:$name");
    debug::log("Added $name") if(DEBUG);

    $cdefid++;
    return @args;
}


sub init {
    my $config = '/etc/parselog/parselog-cgi.conf';

    $conf = read_config_file(file => $config);
    exit 1 unless($conf);

    if ($conf->{debug}) {
        debug->import qw(
            main
            NCGI
            NCGI::Header
            NCGI::Query
            NCGI::Header
        );
        require Log::Delta;
        debug::set_default_log_filter(sub {Log::Delta->instance->l(@_);});
    }

    debug::log("config is $config") if(DEBUG);
    return 1;
}



#
# Lifted and modified from the CPAN ConfigFile module. Has a much
# improved regex.
#
# Version 1.1.0
# Copyright (c) 2002 Sebastien J. Gross. All rights reserved.
# Copyright (c) 2003 Gunnar Wolf. All rights reserved.
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GPL.
#
# Return a hashref from a file, replacing conf if given
#
sub read_config_file {
    my %args = (
        file => '',
        conf => {},
        @_,
    );

    debug::log('read_config_file(file => ',$args{file},'...)') if(DEBUG);

    if (!open(CONF, "< " . $args{file})) {
        warn "Could not open $args{file}: $!\n";
        return undef; 
    }

    my $Conf;
    foreach my $key (keys %{$args{conf}}) {
        $Conf->{$key} = $args{conf}->{$key};
    }

    while(my $line = <CONF>) {
        chomp($line);
        next if($line =~ m/^\s*#/);
        
        $line =~ s/[^\\]#.*$//; # remove inline comments
        $line =~ s/\\#/#/g;     # replace \# with #

        next if $line =~ m/^\s*$/; # ignore empty lines

        # replace $item with $conf->{item}
        $line =~ s{\$(\w+)}{ exists($Conf->{$1}) ? $Conf->{$1} : "\$$1" }gsex;

        if ($line =~ m/\s*(.*[^\s]+)\s*=\s*(.*?)\s*$/) {
            my $s = '$Conf->{\'' . join("\'}->{\'", split /[][]+/, $1) .
                 "\'} = '" . $2 . "'";
            eval $s;
        }
    }
    close CONF;

    return $Conf;
}


#
#
#
sub get_start_end {
    my $period = shift; 

    #
    # Set both $end and $start to midnight (00:00:00) this morning
    #
    my $end = DateTime->now();
    $end->truncate(to => 'day');
    my $start = $end->clone();

    if ($period eq 'yesterday') {
        $start->subtract(days => 1);
    }
    elsif ($period eq 'this_week') {
        $start->truncate(to => 'week');
        $end->truncate(to => 'week');
        $end->add(weeks => 1);
    }
    elsif ($period eq 'last_week') {
        $start->truncate(to => 'week');
        $end->truncate(to => 'week');
        $start->subtract(weeks => 1);
    }
    elsif ($period eq 'this_month') {
        $start->truncate(to => 'month');
        $end->truncate(to => 'month');
        $end->add(months => 1);
    }
    elsif ($period eq 'last_month') {
        $start->truncate(to => 'month');
        $end->truncate(to => 'month');
        $start->subtract(months => 1);
    }
    elsif ($period eq 'this_year') {
        $start->truncate(to => 'year');
        $end->truncate(to => 'year');
        $end->add(years => 1);
    }
    elsif ($period eq 'last_year') {
        $start->truncate(to => 'year');
        $end->truncate(to => 'year');
        $start->subtract(years => 1);
    }
    elsif ($period eq 'two_days') {
        $start->subtract(days => 1);
        $end->add(days => 1);
    }
    else { # return 'today'
        $end->add(days => 1);
    }

    return ($start, $end);
}


__END__


=head1 NAME

parselog.cgi - Generate graphs from parselog RRDs

=head1 SYNOPSIS

  http://hostname/cgi-bin/parselog.cgi

=head1 DESCRIPTION

B<parselog.cgi> is the display part of L<parselog>. It is a Perl CGI
script that will generate a web page with graphs of configuration items.
It is not normally called from the command line, but executed via
a web server.

=head1 CONFIGURATION

/etc/parselog/parselog-cgi.conf is the configuration file. This
file contains settings which determine sizes, time periods and colours
of graphs plus the relative filesystem and URL locations.

All entries in configuration files are of the form 'key = value'.
Lines beginning with '#' and empty lines are ignored. These are the
allowed/required parameters:

=head2 debug

Set this to '1' to turn on debugging. In order for this to work you
also need to have the following modules installed:

  Log::Delta (CPAN)

If you are on debian and have rekudos.net in your sources.list file
you can simply:

  apt-get install liblog-delta-perl

=head2 libdir

Specifies the directory where RRD, state and other permanent files can
be found. Usually /var/lib/parselog.

=head2 imgdir

Specifies the directory where *png files can be created for the
web server to send to clients. Usually /var/cache/parselog-cgi. This
directory must have write permission for the UID that the web server
runs as.

=head2 imgurl

Specifies which URL will map to the B<imgdir> directory.
Usually /parselog-cgi. This depends on some kind of equivalent
configuration in your webserver.

=head2 default_period

Specifies which time period will be used as default. Can be one of
B<today>, B<yesterday>, B<two_days>, B<this_week>, B<last_week>,
B<this_month>, B<last_month>, B<this_year>, B<last_year>.

=head2 width[<period>]

Specifies the number of pixels that will be fed to the rrdgraph
--width parameter for a particular <period>. There should be a
B<width> item for each of B<today>, B<yesterday>, B<two_days>,
B<this_week>, B<last_week>, B<this_month>, B<last_month>,
B<this_year>, B<last_year>.

The value give here is not necessarily what rrdtool will produce.

Example:

  width[today] = 480

=head2 color[<regex>]

Specifies the color (as a six digit hex value) that will be used
in the rrdgraph statement for configuration items. The <regex> may
be any Perl regular expression (see perlre(1) for details).

Example:

  color[Amavis/Blocked INFECTED/.*Netsky] = f0f050

=head1 FILES

/etc/parselog/parselog-cgi.conf
/etc/parselog/css/style.css

=head1 APACHE CONFIGURATION

In order for the web server to be able to find the imgdir directory
and the CSS stylesheet you will need to add something similar to the
following to your apache configuration. See also the conf/apache.conf 
file that comes with the parselog-cgi distribution.

  Alias /parselog-cgi/css /etc/parselog/css/
  <Directory /etc/parselog/css>
    AllowOverride None
    Order allow,deny
    Allow from all
  </Directory>

  Alias /parselog-cgi/ /var/cache/parselog-cgi/
  <Directory /var/cache/parselog-cgi>
    AllowOverride None
    Order allow,deny
    Allow from all
  </Directory>

=head1 SEE ALSO

L<parselog>(1), L<NCGI>, http://rekudos.net/parselog/

=head1 AUTHOR

Mark Lawrence E<lt>nomad@null.netE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright (C) 2004,2005 Mark Lawrence <nomad@null.net>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

=cut

